 // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ //.
//  OpenGL for FPC - Tutorial 16                      //
//                                                    //
//  A simple object, now composed with display lists. //
//                                                    //
//  Works with FPC 1.0.10.                            //
//                                                    //
//  (c)2002-2003 Delax^Sundancer Inc.                 //
//   delax@sundancerinc.de                            //
//                                                    //
//   www.friends-of-fpc.org                           //
 // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ //'

{$APPTYPE GUI}
{$MODE DELPHI}

program ogl_fpc_source16;

uses
  dos,SysUtils,windows,math,CommDlg,CommCtrl,gl,glu,glext;

var

  msg_ : MSG;		// Windows messages
  hWindow : HWnd;	// Windows Handle to the OGL Window
  dcWindow : hDc;	// Device Context for the OGL Window
  rcWindow : HGLRC;	// Render Context for the OGL Window
 
  width, height, bits : longint;
  fullscreen, active : boolean;

  rotation : real;

  AmbientLight : array[0..3] of glFloat = (0.5,0.5,0.5,1.0);
  DiffuseLight : array[0..3] of glFloat = (1.0,1.0,1.0,1.0);
  SpecularLight : array[0..3] of glFloat = (1.0,1.0,1.0,1.0);
  SpecularReflection : array[0..3] of glFloat = (0.4,0.4,0.4,1.0);
  LightPosition : array[0..3] of glFloat;

  CNormal : array[1..3] of real;

  Pillar, Ground : integer;
  LightX, LightY, LightZ : real;
  LightMulX, LightMulY, LightMulZ : real;

// Try, Throw, Catch mechanism. Simple proc to display given errors. //
procedure ThrowError(pcErrorMessage : pChar);
begin
  MessageBox(0, pcErrorMessage, 'Error', MB_OK);
  Halt(0);
end;


// Standard WinProc. Handles all the messages from the System. //
function GLWndProc(Window: HWnd; AMessage, WParam, LParam: Longint): Longint; stdcall; export;
begin
  GLWndProc := 0;

  case AMessage of
    wm_create:
      begin
	active := true;		// if GL Window was created correctly, then set
	Exit;			// active-flag to "true".
      end;

    wm_paint:
      begin
         exit;			// nothing to paint to Windows as we do all drawing with OGL
      end;

    wm_keydown:
      begin
	if wParam = VK_ESCAPE then SendMessage(hWindow,wm_destroy,0,0);
	Exit;			// check for ESC key. If pressed, then send quit message
      end;

    wm_destroy:
      begin
         active := false;	// if quit message was sent, exit the main loop by setting
         PostQuitMessage(0);	// the active-flag to "false".
         Exit;
      end;

   wm_syscommand:		// system wants something..
      begin
	case (wParam) of 
	  SC_SCREENSAVE : begin		// ..don't start any screensavers.
	        GLWndProc := 0;
	       end;

	  SC_MONITORPOWER : begin	// ..and don't kill monitor power.
	        GLWndProc := 0;
	       end;
        end;
      end;
 end;

  GLWndProc := DefWindowProc(Window, AMessage, WParam, LParam);	// let Windows deal with the rest of the messages.
end;


// Register the Window Class. //
function WindowRegister: Boolean;
var
  WindowClass: WndClass;
begin
  WindowClass.Style := cs_hRedraw or cs_vRedraw;
  WindowClass.lpfnWndProc := WndProc(@GLWndProc);		// Handle to our Windows messaging interface func.
  WindowClass.cbClsExtra := 0;
  WindowClass.cbWndExtra := 0;
  WindowClass.hInstance := system.MainInstance;			// Get the Windows Instance for our app. 
  WindowClass.hIcon := LoadIcon(0, idi_Application);
  WindowClass.hCursor := LoadCursor(0, idc_Arrow);
  WindowClass.hbrBackground := GetStockObject(WHITE_BRUSH);
  WindowClass.lpszMenuName := nil;
  WindowClass.lpszClassName := 'GLWindow';			// Name the specified Window Class

  WindowRegister := RegisterClass(WindowClass) <> 0;
end;


// Create the OGL Window. //
function WindowCreate(pcApplicationName : pChar): HWnd;
var
  hWindow: HWnd;		// Handle to Window
  dmScreenSettings : DEVMODE;	// Used for Full Screen Mode
begin

 if fullscreen = false then begin	// OGL Window running in windowed mode

  hWindow := CreateWindow('GLWindow',
			  pcApplicationName,
			  WS_CAPTION OR WS_POPUPWINDOW OR WS_VISIBLE OR WS_CLIPSIBLINGS OR WS_CLIPCHILDREN,
			  cw_UseDefault,
			  cw_UseDefault,
			  width,
			  height,
			  0, 0,
			  system.MainInstance,
			  nil);

 end else begin		// OGL Window running in fullscreen mode

  dmScreenSettings.dmSize := sizeof(dmScreenSettings);		// Size Of The Devmode Structure
  dmScreenSettings.dmPelsWidth := width;		// Selected Screen Width
  dmScreenSettings.dmPelsHeight := height;		// Selected Screen Height
  dmScreenSettings.dmBitsPerPel := bits;		// Selected Bits Per Pixel
  dmScreenSettings.dmFields := DM_BITSPERPEL OR DM_PELSWIDTH OR DM_PELSHEIGHT;

 if ChangeDisplaySettings(@dmScreenSettings,CDS_FULLSCREEN) <> DISP_CHANGE_SUCCESSFUL then begin
    ThrowError('Screen resolution is not supported by your gfx card!');
    WindowCreate := 0;
    Exit;
 end;

  hWindow := CreateWindowEx(WS_EX_APPWINDOW,
			    'GLWindow',
			    pcApplicationName,
			    WS_POPUP OR WS_VISIBLE OR WS_CLIPSIBLINGS OR WS_CLIPCHILDREN,
			    0, 0,
			    width,
			    height,
			    0, 0,
			    system.MainInstance,
			    nil );

  ShowCursor(false);
 end;

  if hWindow <> 0 then begin
    ShowWindow(hWindow, CmdShow);
    UpdateWindow(hWindow);
  end;

  WindowCreate := hWindow;
end;


// Init the Window and bind OGL to it. //
function WindowInit(hParent : HWnd): Boolean;
var
  FunctionError : integer;
  pfd : PIXELFORMATDESCRIPTOR;
  iFormat : integer;	// Pixel Format

begin

 FunctionError := 0;

  dcWindow := GetDC( hParent );	// Get Device Context

  FillChar(pfd, sizeof(pfd), 0);	// Define Pixel Format
  pfd.nSize         := sizeof(pfd);
  pfd.nVersion      := 1;
  pfd.dwFlags       := PFD_SUPPORT_OPENGL OR PFD_DRAW_TO_WINDOW OR PFD_DOUBLEBUFFER or PFD_GENERIC_FORMAT or PFD_GENERIC_ACCELERATED;
  pfd.iPixelType    := PFD_TYPE_RGBA;
  pfd.cColorBits    := bits;
  pfd.cDepthBits    := 16;
  pfd.iLayerType    := PFD_MAIN_PLANE;

  iFormat := ChoosePixelFormat( dcWindow, @pfd );	// Create Pixel Format

  if (iFormat = 0) then FunctionError := 1;

  SetPixelFormat( dcWindow, iFormat, @pfd );		// Set Pixel Format
  rcWindow := wglCreateContext( dcWindow );		// Create OpenGL Context

  if (rcWindow = 0) then FunctionError := 2;

  wglMakeCurrent( dcWindow, rcWindow );		// Bind OpenGL to our Window

 if FunctionError = 0 then WindowInit := true else WindowInit := false;

end;


// Main function to create the Window. //
function CreateOGLWindow(pcApplicationName : pChar; iApplicationWidth, iApplicationHeight, iApplicationBits : longint; bApplicationFullscreen : boolean):Boolean;
begin
 width := iApplicationWidth;
 height := iApplicationHeight;
 bits := iApplicationBits;
 fullscreen := bApplicationFullscreen;

  if not WindowRegister then begin
    ThrowError('Could not register the Application Window!');
    CreateOGLWindow := false;
    Exit;
  end;

  hWindow := WindowCreate(pcApplicationName);
  if longint(hWindow) = 0 then begin
    ThrowError('Could not create Application Window!');
    CreateOGLWindow := false;
    Exit;
  end;

 if not WindowInit(hWindow) then begin
    ThrowError('Could not initialise Application Window!');
    CreateOGLWindow := false;
    Exit;
  end;

 CreateOGLWindow := true;
end;


// Kill Application Window again. //
procedure KillOGLWindow;
begin
  wglMakeCurrent( dcWindow, 0 );	// Kill Device Context
  wglDeleteContext( rcWindow );		// Kill Render Context
  ReleaseDC( hWindow, dcWindow );	// Release Window
  DestroyWindow( hWindow );		// Kill Window itself
end;


// Find normal of a surface. //
procedure FindNormal(v1x, v1y, v1z, v2x, v2y, v2z, v3x, v3y, v3z : real);
const	x = 1;
	y = 2;
	z = 3;
var	temp_v1, temp_v2 : array[1..3] of real;
	temp_lenght : real;
begin

temp_v1[x] := v1x - v2x;
temp_v1[y] := v1y - v2y;
temp_v1[z] := v1z - v2z;

temp_v2[x] := v2x - v3x;
temp_v2[y] := v2y - v3y;
temp_v2[z] := v2z - v3z;

// calculate cross product
CNormal[x] := temp_v1[y]*temp_v2[z] - temp_v1[z]*temp_v2[y];
CNormal[y] := temp_v1[z]*temp_v2[x] - temp_v1[x]*temp_v2[z];
CNormal[z] := temp_v1[x]*temp_v2[y] - temp_v1[y]*temp_v2[x];

// normalize normal
temp_lenght :=	(CNormal[x]*CNormal[x])+
		(CNormal[y]*CNormal[y])+
		(CNormal[z]*CNormal[z]);

temp_lenght := sqrt(temp_lenght);

// prevent n/0
if temp_lenght = 0 then temp_lenght := 1;

CNormal[x] := CNormal[x]/temp_lenght;
CNormal[y] := CNormal[y]/temp_lenght;
CNormal[z] := CNormal[z]/temp_lenght;

end;


// OGL initialisations //
procedure OpenGL_Init();
begin

  glClearColor( 0.0, 0.0, 0.0, 0.0 );

  glViewport( 0, 0, width, height );
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  gluPerspective(45.0,width/height,0.1,100.0);

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  glClearDepth(1.0);               // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);         // Enables Depth Testing
  glDepthFunc(GL_LEQUAL);          // The Type Of Depth Test To Do

  glEnable(GL_CULL_FACE);          // Enable Hidden Surface Removal
  glCullFace(GL_BACK);             // Set to Back
  glFrontFace(GL_CCW);             // Draw all surfaces CCW

  glShadeModel(GL_SMOOTH);         // Set shading model

  glEnable(GL_LIGHTING);                            // Enable Lighting
  glLightfv(GL_LIGHT0,GL_AMBIENT, @AmbientLight);   // Enable Ambient Light
  glLightfv(GL_LIGHT0,GL_DIFFUSE, @DiffuseLight);   // Enable Diffuse Light
  glLightfv(GL_LIGHT0,GL_SPECULAR, @SpecularLight); // Enable Specular Light
  glEnable(GL_LIGHT0);                              // Enable Light0

  LightPosition[0] := 0.0;  LightPosition[1] := 0.0;
  LightPosition[2] := 0.0;  LightPosition[3] := 1.0;

  glLightfv(GL_LIGHT0,GL_POSITION,@LightPosition);   // Set our Light Position

  glEnable(GL_COLOR_MATERIAL);                       // Enable material colors
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE); // Set to front for Ambient and Diffuse Light
  glMaterialfv(GL_FRONT, GL_SPECULAR, @SpecularReflection); // Set reflection
  glMateriali(GL_FRONT,GL_SHININESS,10);             // Set shininess of materials

end;


procedure OpenGL_CreateList();
begin

 Pillar := glGenLists(1);
 glNewList(Pillar,GL_COMPILE);

  glBegin(GL_QUADS);

   glNormal3f(1.0, 0.0, 0.0);
    glVertex3f( 1.0, 1.0,-1.0);
    glVertex3f( 1.0, 1.0, 1.0); // rechts unten
    glVertex3f( 1.0,-1.0, 1.0);
    glVertex3f( 1.0,-1.0,-1.0);

   glNormal3f(-1.0, 0.0, 0.0);
    glVertex3f(-1.0, 1.0, 1.0);
    glVertex3f(-1.0, 1.0,-1.0); // links unten
    glVertex3f(-1.0,-1.0,-1.0);
    glVertex3f(-1.0,-1.0, 1.0);

   glNormal3f(0.0,1.0,0.0);
    glVertex3f( 1.0, 1.0, 1.0);
    glVertex3f( 1.0, 1.0,-1.0); // hinten unten
    glVertex3f(-1.0, 1.0,-1.0);
    glVertex3f(-1.0, 1.0, 1.0);

   glNormal3f(0.0, -1.0, 0.0);
    glVertex3f( 1.0,-1.0,-1.0);
    glVertex3f( 1.0,-1.0, 1.0); // vorne unten
    glVertex3f(-1.0,-1.0, 1.0);
    glVertex3f(-1.0,-1.0,-1.0);

   FindNormal(-0.5,0.5,6.0,0.5,0.5,6.0,1.0,1.0,1.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f(-0.5, 0.5, 6.0);
    glVertex3f( 0.5, 0.5, 6.0); // vorne oben
    glVertex3f( 1.0, 1.0, 1.0);
    glVertex3f(-1.0, 1.0, 1.0);

   FindNormal(0.5,-0.5,6.0,-0.5,-0.5,6.0,-1.0,-1.0,1.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f( 0.5,-0.5, 6.0);
    glVertex3f(-0.5,-0.5, 6.0); // hinten oben
    glVertex3f(-1.0,-1.0, 1.0);
    glVertex3f( 1.0,-1.0, 1.0);

   FindNormal(1.0,1.0,1.0,0.5,0.5,6.0,0.5,-0.5,6.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f( 1.0, 1.0, 1.0);
    glVertex3f( 0.5, 0.5, 6.0); // rechts oben
    glVertex3f( 0.5,-0.5, 6.0);
    glVertex3f( 1.0,-1.0, 1.0);

   FindNormal(-0.5,0.5,6.0,-1.0,1.0,1.0,-1.0,-1.0,1.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f(-0.5, 0.5, 6.0);
    glVertex3f(-1.0, 1.0, 1.0); // links oben
    glVertex3f(-1.0,-1.0, 1.0);
    glVertex3f(-0.5,-0.5, 6.0);

glEnd(); 

glBegin(GL_TRIANGLES);

   FindNormal(-0.5,0.5,6.0,0.0,0.0,7.0,0.5,0.5,6.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f(-0.5, 0.5, 6.0);
    glVertex3f( 0.0, 0.0, 7.0);
    glVertex3f( 0.5, 0.5, 6.0); // vorne oben

   FindNormal(0.5,-0.5,6.0,0.0,0.0,7.0,-0.5,-0.5,6.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f( 0.5,-0.5, 6.0);
    glVertex3f( 0.0, 0.0, 7.0);
    glVertex3f(-0.5,-0.5, 6.0); // hinten oben

   FindNormal(0.5,0.5,6.0,0.0,0.0,7.0,0.5,-0.5,6.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f( 0.5, 0.5, 6.0); // rechts oben
    glVertex3f( 0.0, 0.0, 7.0);
    glVertex3f( 0.5,-0.5, 6.0);

   FindNormal(-0.5,-0.5,6.0,0.0,0.0,7.0,-0.5,0.5,6.0);
   glNormal3f(CNormal[1], CNormal[2], CNormal[3]);
    glVertex3f(-0.5,-0.5, 6.0);
    glVertex3f( 0.0, 0.0, 7.0);
    glVertex3f(-0.5, 0.5, 6.0); // links oben

glEnd;

glEndList();

 Ground := glGenLists(1);
 glNewList(Ground,GL_COMPILE);
  glBegin(GL_QUADS);

   glNormal3f(0.0,0.0,1.0);
    glVertex3f( 10.0,-10.0, 0.0);
    glVertex3f( 10.0, 10.0, 0.0);
    glVertex3f(-10.0, 10.0, 0.0);
    glVertex3f(-10.0,-10.0, 0.0);
  glEnd(); 
glEndList();

end;


procedure OpenGL_Draw();
begin

  glClear(GL_COLOR_BUFFER_BIT OR GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();

  glTranslatef( 0.0,-5.0,-40.0);
  glRotatef( 80,-1.0, 0.0, 0.0);
  glRotatef( rotation, 0.0, 0.0, 1.0);

  LightPosition[0] := LightX;  LightPosition[1] := LightY;
  LightPosition[2] := LightZ;  LightPosition[3] := 0.7;
  glLightfv(GL_LIGHT0,GL_POSITION,@LightPosition);   // Set our Light Position

  glColor3f( 1.0, 1.0, 1.0 );
  glBegin (GL_POINTS);
  glNormal3f(0.0,0.0, 1.0);
   glVertex3f(LightPosition[0],LightPosition[1],LightPosition[2]);
  glEnd();


  glColor3f( 0.0, 0.0, 0.5 );
  glTranslatef( 0.0, 0.0, 1.0);
  glCallList(Ground);

  glColor3f( 0.9, 0.0, 0.0 );
  glTranslatef( 5.0,5.0,0.0);
  glCallList(Pillar);

  glColor3f( 0.0, 0.9, 0.0 );
  glTranslatef( 0.0,-10.0,0.0);
  glCallList(Pillar);

  glColor3f( 0.9, 0.9, 0.0 );
  glTranslatef(-10.0,0.0,0.0);
  glCallList(Pillar);

  glColor3f( 0.0, 0.9, 0.9 );
  glTranslatef(0.0,10.0,0.0);
  glCallList(Pillar);

end;


// Main Loop. //
begin

 if MessageBox(0,'Fullscreen Mode?', 'Question!',MB_YESNO OR MB_ICONQUESTION) = IDNO 
  then fullscreen := false else fullscreen := true;

 CreateOGLWindow('Source 16', 640, 480, 32, fullscreen);
 OpenGL_Init();			// init opengl stuff
 OpenGL_CreateList();

 rotation := 0.0;

 LightX := 0.0;
 LightY := 0.0;
 LightZ := 3.0;

 LightMulX := (random(9)+1) / 200;
 LightMulY := (random(9)+1) / 200;
 LightMulZ := (random(9)+1) / 200;

 REPEAT				// start main proc

if PeekMessage(@msg_,0,0,0,0) = true then begin
  GetMessage(@msg_,0,0,0);
  TranslateMessage(msg_);
  DispatchMessage(msg_);
end;

 OpenGL_Draw();
 SwapBuffers( dcWindow );	// put opengl stuff to screen

 rotation := rotation+0.2;

 if (LightX < -10) or (LightX > 10) then LightMulX := LightMulX*-1;
 if (LightY < -10) or (LightY > 10) then LightMulY := LightMulY*-1;
 if (LightZ < 1) or (LightZ > 20) then LightMulZ := LightMulZ*-1;

 LightX := LightX+LightMulX;
 LightY := LightY+LightMulY;
 LightZ := LightZ+LightMulZ;

UNTIL active = false;		// end main proc

  KillOGLWindow();		// kill window stuff
end.
